# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import functools
import importlib.util

from . import NoData, BaseClass, MissingImport
from .help import FullManPage, TypesManPage, PluginManPage, PluginsManPage, \
    AttributesManPage
from .type import Type
from .cache import NODATA, NOTSET
from .plugin import Speed, Plugin
from .attribute import Attribute
from .exceptions import UsageError, BadPluginError, BadAttributeError


class Registry(BaseClass):
    """Registry class that keeps track of all installed plugins and the attributes they provide.
    """
    # pylint:disable=too-many-public-methods

    PLUGIN_DIR_BUILTIN = os.path.join(os.path.dirname(__file__), "builtin")
    PLUGIN_DIR_CONTRIB = "/usr/lib/ff"
    PLUGIN_DIR_USER = os.path.expanduser("~/.ff")

    registered_types = set()
    registered_plugins = {}

    plugins = {}
    attributes = {}

    def __init__(self, context):
        super().__init__(context)

        self.optimize_for_slow_plugins = False

        with self.context.global_lock:
            if not self.registered_plugins:
                self.load_plugins()

    def iter_plugin_dirs(self):
        """Yield possible plugin directories.
        """
        yield "builtin", self.PLUGIN_DIR_BUILTIN

        # Return different plugin directories depending on whether we are in the git working tree.
        dirname = os.path.dirname(__file__)
        git_dir = os.path.join(dirname, "../.git")
        plugin_dir = os.path.join(dirname, "../plugins")
        if __debug__ and os.path.isdir(git_dir) and os.path.isdir(plugin_dir):
            yield "contrib", os.path.realpath(plugin_dir)
        else:
            yield "contrib", self.PLUGIN_DIR_CONTRIB

        plugin_dirs = os.environ.get("FF_PLUGIN_DIRS")
        if plugin_dirs is not None:
            for directory in plugin_dirs.split(os.pathsep):
                if directory:
                    yield "system", directory

        yield "user", self.PLUGIN_DIR_USER

    def load_plugin(self, name, source, path):
        """Load the plugin code from path into a python module. After that inpect the module
           namespace to find the Type and Plugin classes.
        """
        spec = importlib.util.spec_from_file_location(name, path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        self.inspect_namespace(source, path, module)

    def inspect_namespace(self, source, path, module):
        """Go through the objects in a module namespace and collect Type and Plugin classes.
        """
        # pylint:disable=unidiomatic-typecheck
        for obj in vars(module).values():
            if type(obj) is not type:
                continue

            if issubclass(obj, Type) and obj is not Type:
                self.register_type(obj)

            elif issubclass(obj, Plugin) and obj is not Plugin:
                self.register_plugin(source, path, obj)

    def register_type(self, type_cls):
        """Register Type classes.
        """
        self.registered_types.add(type_cls)

    def register_plugin(self, source, path, plugin_cls):
        """Register and initialize Plugin classes.
        """
        if plugin_cls.name in self.registered_plugins:
            self.logger.warning(f"Skipping already loaded plugin {plugin_cls.name!r} "\
                    f"from {path!r}")
            return

        plugin_cls.initialize(self.logger, source, path)

        self.registered_plugins[plugin_cls.name] = plugin_cls

        for name, type_cls, _ in plugin_cls.attributes:
            if not issubclass(type_cls, tuple(self.registered_types)):
                raise BadPluginError.from_exception(f"plugin {plugin_cls.name!r} attribute "\
                        f"{name!r} uses invalid type {type_cls!r}")

    def load_plugins(self):
        """Load all available plugins from the designated plugin directories.
        """
        for source, directory in self.iter_plugin_dirs():
            try:
                with os.scandir(directory) as direntries:
                    for direntry in direntries:
                        if not direntry.is_file() or not direntry.name.endswith(".py") or \
                                direntry.name.startswith("_"):
                            continue

                        try:
                            self.load_plugin(direntry.name[:-3], source, direntry.path)
                        except Exception as exc:
                            raise BadPluginError.from_exception(
                                    f"Plugin file {direntry.path!r} failed to load") from exc

            except OSError:
                continue

        # Sort file attributes first.
        for name, plugin_cls in sorted(self.registered_plugins.items(),
                key=lambda t: "" if t[0] == "file" else t[0]):
            for attr, type_cls, _ in sorted(plugin_cls.attributes):
                self.attributes[Attribute(name, attr)] = type_cls

    def init_plugin(self, name):
        """Setup the Plugin class for later use (register it in the cache and call the setup()
           method).
        """
        # pylint:disable=broad-except
        plugin_cls = self.registered_plugins[name]
        try:
            if plugin_cls.use_cache:
                self.cache.register_plugin(plugin_cls)
            self.plugins[name] = plugin_cls()
        except MissingImport as exc:
            raise BadPluginError(f"Plugin {name!r} needs the following module: {exc.module!r}") \
                    from exc
        except Exception as exc:
            raise BadPluginError.from_exception(f"Unable to setup plugin {name!r}") from exc

        if plugin_cls.speed is Speed.SLOW:
            # See FilesystemWalker for an explanation.
            self.optimize_for_slow_plugins = True

    def parse_attribute(self, name):
        """Break down an attribute into plugin name and attribute name. If the attribute is without
           a plugin name guess which one it is.
        """
        try:
            attribute = Attribute(*name.split(".", 1))
        except TypeError as exc:
            plugin_names = self.get_plugin_for_attribute(name)

            # The file plugin always has precedence. This way we avoid
            # essential attributes being overwritten by rogue plugins.
            if "file" in plugin_names:
                attribute = Attribute("file", name, name)

            elif len(plugin_names) == 1:
                # The attribute is provided by exactly one plugin.
                attribute = Attribute(plugin_names.pop(), name, name)

            elif not plugin_names:
                # We could not find the attribute in any of the plugins.
                raise BadAttributeError(f"No plugin found for attribute {name!r}") from exc

            else:
                # There is more than one plugin providing this attribute, we
                # cannot continue.
                raise BadAttributeError(f"Attribute {name!r} is ambiguous (choose between " \
                        f"{', '.join(f'{plugin_name}.{name}' for plugin_name in plugin_names)})") \
                        from exc

        return attribute

    def setup_attribute(self, attribute):
        """Setup an attribute for later use, i.e. setup the plugin that provides it if it has not
           been setup already.
        """
        attribute = self.parse_attribute(attribute)

        if attribute.plugin not in self.registered_plugins:
            raise BadAttributeError(f"No such plugin {attribute.plugin!r}")

        with self.context.global_lock:
            if attribute.plugin not in self.plugins:
                self.init_plugin(attribute.plugin)

        return attribute

    @functools.lru_cache(maxsize=128)
    def get_data(self, entry, plugin):
        """Return all attribute values that are associated with entry.
        """
        plugin = self.plugins[plugin]
        try:
            return self.get_data_from_plugin(entry, plugin)
        except NoData:
            return {}
        except NotImplementedError as exc:
            raise BadPluginError.from_exception(
                    f"Plugin {plugin.name!r} is not completely implemented") from exc
        except Exception as exc:
            raise BadPluginError.from_exception(
                    f"Plugin {plugin.name!r} had an unhandled exception") from exc

    def get_data_from_plugin(self, entry, plugin):
        """Let the plugin process the entry.
        """
        data = {}

        if plugin.can_handle(entry):
            if plugin.use_cache:
                tag = plugin.get_entry_cache_tag(entry)
                cached = self.cache.get(plugin, entry.abspath, tag)
                if cached is NOTSET:
                    # There is no cached result for this entry, so we ask the plugin for the data.
                    # Even if the plugin fails to process this entry, we cache the return value, so
                    # that we know we don't have to try again in the future.
                    if __debug__:
                        self.logger.debug("cache",
                                f"Cache {plugin.name!r} data for {entry.path!r}")

                    try:
                        cached = plugin.cache(entry)
                    except NoData:
                        # Do not bother to call process() below, we could not handle this entry.
                        cached = NODATA

                    self.cache.set(plugin, entry.abspath, tag, cached)

            else:
                cached = None

            if cached is not NODATA:
                data = dict(plugin.process(entry, cached))

        if __debug__:
            for key, value in data.items():
                type_cls = self.get_attribute_type(Attribute(plugin.name, key))
                if not type_cls.check_type(value):
                    raise BadPluginError(
                        f"plugin {plugin.name!r} produced value {key}={value!r} that is "\
                        f"not of type {type_cls.name!r}")

        return data

    def get_attribute(self, entry, attribute):
        """Return the attribute value that is associated with entry.
        """
        if attribute.plugin == "file":
            # Use a shortcut for attributes from the 'file' plugin. Fetch the value directly from
            # the Entry object instead of going through the plugin.
            return entry.get_attribute(attribute.name)

        data = self.get_data(entry, attribute.plugin)
        return data[attribute.name]

    def get_attribute_and_type(self, entry, attribute):
        """Return the attribute value and its type.
        """
        return self.get_attribute(entry, attribute), self.attributes[attribute]

    def get_attribute_type(self, attribute):
        """Return the Type associated with the attribute.
        """
        try:
            return self.attributes[attribute]
        except KeyError as exc:
            raise BadAttributeError(
                    f"{attribute.plugin!r} plugin has no attribute {attribute.name!r}") from exc

    def get_plugin_speed(self, attribute):
        """Return the speed score of the plugin associated with the attribute.
        """
        return self.plugins[attribute.plugin].speed.value

    def get_file_attributes(self):
        """Get all attributes associated with the 'file' plugin.
        """
        for attribute in self.attributes:
            if attribute.plugin == "file":
                yield attribute.name

    def get_plugin_for_attribute(self, name):
        """Return a set of possible plugin names for an attribute name.
        """
        return set(attribute.plugin for attribute in self.attributes if attribute.name == name)

    def get_full_manpage(self):
        """Return the full help a.k.a. the manpage.
        """
        return FullManPage()

    def get_plugin_manpage(self, name):
        """Return the help for a specific plugin.
        """
        try:
            plugin = self.registered_plugins[name]
        except KeyError as exc:
            raise UsageError(f"Plugin {name!r} not found") from exc

        return PluginManPage(plugin)

    def get_attributes_manpage(self):
        """Print a table of all available attributes.
        """
        plugins = list(self.registered_plugins.values())
        plugins.sort(key=lambda p: "" if p.name == "file" else p.name)
        return AttributesManPage(plugins)

    def get_plugins_manpage(self):
        """Reutrn the help for all available plugins.
        """
        return PluginsManPage(sorted(self.registered_plugins.values(), key=lambda p: p.name))

    def get_types_manpage(self):
        """Return the help for available types.
        """
        return TypesManPage(sorted(self.registered_types, key=lambda t: t.name))



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import itertools

from . import BaseClass
from .parser import Parser


class MatcherParser(Parser):
    """The MatcherParser warns about using the file.hide/file.hidden attributes without using
       -H/--hidden.
    """

    def create_test(self, attribute, operator, reference, value):
        if self.args.hide and attribute.plugin == "file" and \
                attribute.name in ("hide", "hidden"):
            self.logger.warning("file.hide and file.hidden have no effect with -H/--hide",
                    tag="hide")
        return super().create_test(attribute, operator, reference, value)


class Filter(BaseClass):
    """Basic Filter class to build Excluder and Matcher upon.
    """

    def __init__(self, context, parser):
        super().__init__(context)
        self.parser = parser

        if self.is_empty():
            self.test = self.test_empty

    def is_empty(self):
        """Return True if the test sequence contains no tests.
        """
        return not list(itertools.chain.from_iterable(self.parser))


class Excluder(Filter):
    """Excluder class that allows to exclude Entry objects based on their attributes.

       The Excluder works differently from the Matcher. It is simpler and only supports a flat list
       of tests.
    """
    # pylint:disable=method-hidden

    def __init__(self, context, tokens):
        super().__init__(context, Parser(context, tokens))

    def test_empty(self, entry):
        """Default test method when the list of tests is empty.
        """
        # pylint:disable=unused-argument
        return False

    def test(self, entry):
        """Evaluate the list of Test objects for the entry object. Return True if the entry matches
           and is to be excluded.
        """
        for test in itertools.chain.from_iterable(self.parser):
            try:
                value, type_cls = self.registry.get_attribute_and_type(entry, test.attribute)
            except KeyError:
                # Ignore entries that don't provide this attribute.
                continue

            # If an entry matches the test, we return immediately without further going through the
            # TestSequence.
            if type_cls.test(entry, test, value):
                return True

        return False


class Matcher(Filter):
    """Matcher class that allows filtering Entry objects based on their attributes.
    """
    # pylint:disable=method-hidden

    def __init__(self, context, tokens):
        super().__init__(context, MatcherParser(context, tokens))

    def test_empty(self, entry):
        """Default test method when the list of tests is empty.
        """
        # pylint:disable=unused-argument
        return True

    def test(self, entry):
        """Evaluate the main sequence of Test objects and its sub sequences for the attributes from
           the entry object. Return True if the entry matches.
        """
        return any(self._test(self.parser, entry))

    def _test(self, tests, entry):
        """Handle a single sub test sequence by evaluating the Test objects it contains.
        """
        for test in tests:
            if test.name == "AND":
                yield all(self._test(test, entry))

            elif test.name == "OR":
                yield any(self._test(test, entry))

            elif test.name == "NOT":
                yield not any(self._test(test, entry))

            else:
                # Fetch the attribute's value from the entry to test.
                try:
                    value, type_cls = self.registry.get_attribute_and_type(entry, test.attribute)
                except KeyError:
                    # The plugin was unable to provide a value for this attribute, so we remove
                    # this entry from the result.
                    yield False
                else:
                    yield type_cls.test(entry, test, value)



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import sys

from libff.plugin import *


class Xattr(Plugin):
    """The "xattr" plugin provides access to a file's extended attributes.
    """

    attributes = [
        ("xattrs",  ListOfStrings, "A list of key=value pairs from the file's extended attributes.")
    ]

    def can_handle(self, entry):
        return entry.is_dir() or entry.is_file()

    def process(self, entry, cached):
        xattrs = []
        try:
            for key in os.listxattr(entry.path):
                value = os.getxattr(entry.path, key)
                try:
                    value = os.fsdecode(value)
                except UnicodeDecodeError:
                    value = value.decode(sys.getfilesystemencoding(), "backslashreplace")
                xattrs.append(f"{key}={value}")
        except OSError:
            pass

        yield "xattrs", xattrs



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
from ctypes import *

from libff.plugin import *

libc = CDLL("libc.so.6", use_errno=True)

fs_types = {
    # Lifted from src/stat.c from coreutils.
    0x61636673: ("acfs", True), 0xADF5: ("adfs", False), 0xADFF: ("affs", False),
    0x5346414F: ("afs", True), 0x09041934: ("anon-inode FS", False), 0x61756673: ("aufs", True),
    0x0187: ("autofs", False), 0x42465331: ("befs", False), 0x62646576: ("bdevfs", False),
    0x1BADFACE: ("bfs", False), 0xCAFE4A11: ("bpf_fs", False), 0x42494E4D: ("binfmt_misc", False),
    0x9123683E: ("btrfs", False), 0x73727279: ("btrfs_test", False), 0x00C36400: ("ceph", True),
    0x0027E0EB: ("cgroupfs", False), 0xFF534D42: ("cifs", True), 0x73757245: ("coda", True),
    0x012FF7B7: ("coh", False), 0x62656570: ("configfs", False), 0x28CD3D45: ("cramfs", False),
    0x453DCD28: ("cramfs-wend", False), 0x64626720: ("debugfs", False), 0x1373: ("devfs", False),
    0x1CD1: ("devpts", False), 0xF15F: ("ecryptfs", False), 0xDE5E81E4: ("efivarfs", False),
    0x00414A53: ("efs", False), 0x5DF5: ("exofs", False), 0x137D: ("ext", False),
    0xEF53: ("ext2/ext3", False), 0xEF51: ("ext2", False), 0xF2F52010: ("f2fs", False),
    0x4006: ("fat", False), 0x19830326: ("fhgfs", True), 0x65735546: ("fuseblk", True),
    0x65735543: ("fusectl", True), 0x0BAD1DEA: ("futexfs", False), 0x01161970: ("gfs/gfs2", True),
    0x47504653: ("gpfs", True), 0x4244: ("hfs", False), 0x482B: ("hfs+", False),
    0x4858: ("hfsx", False), 0x00C0FFEE: ("hostfs", False), 0xF995E849: ("hpfs", False),
    0x958458F6: ("hugetlbfs", False), 0x11307854: ("inodefs", False), 0x013111A8: ("ibrix", True),
    0x2BAD1DEA: ("inotifyfs", False), 0x9660: ("isofs", False), 0x4004: ("isofs", False),
    0x4000: ("isofs", False), 0x07C0: ("jffs", False), 0x72B6: ("jffs2", False),
    0x3153464A: ("jfs", False), 0x6B414653: ("k-afs", True), 0xC97E8168: ("logfs", False),
    0x0BD00BD0: ("lustre", True), 0x137F: ("minix", False), 0x138F: ("minix (30 char.)", False),
    0x2468: ("minix v2", False), 0x2478: ("minix v2 (30 char.)", False), 0x4D5A: ("minix3", False),
    0x19800202: ("mqueue", False), 0x4D44: ("msdos", False), 0x564C: ("novell", True),
    0x6969: ("nfs", True), 0x6E667364: ("nfsd", True), 0x3434: ("nilfs", False),
    0x6E736673: ("nsfs", False), 0x5346544E: ("ntfs", False), 0x9FA1: ("openprom", False),
    0x7461636F: ("ocfs2", True), 0x794C7630: ("overlayfs", True), 0xAAD7AAEA: ("panfs", True),
    0x50495045: ("pipefs", True), 0x9FA0: ("proc", False), 0x6165676C: ("pstorefs", False),
    0x002F: ("qnx4", False), 0x68191122: ("qnx6", False), 0x858458F6: ("ramfs", False),
    0x52654973: ("reiserfs", False), 0x7275: ("romfs", False), 0x67596969: ("rpc_pipefs", False),
    0x73636673: ("securityfs", False), 0xF97CFF8C: ("selinux", False),
    0x43415D53: ("smackfs", False), 0x517B: ("smb", True), 0xBEEFDEAD: ("snfs", True),
    0x534F434B: ("sockfs", False), 0x73717368: ("squashfs", False), 0x62656572: ("sysfs", False),
    0x012FF7B6: ("sysv2", False), 0x012FF7B5: ("sysv4", False), 0x01021994: ("tmpfs", False),
    0x74726163: ("tracefs", False), 0x24051905: ("ubifs", False), 0x15013346: ("udf", False),
    0x00011954: ("ufs", False), 0x54190100: ("ufs", False), 0x9FA2: ("usbdevfs", False),
    0x01021997: ("v9fs", False), 0xBACBACBC: ("vmhgfs", True), 0xA501FCF5: ("vxfs", True),
    0x565A4653: ("vzfs", False), 0xABBA1974: ("xenfs", False), 0x012FF7B4: ("xenix", False),
    0x58465342: ("xfs", False), 0x012FD16D: ("xia", False), 0x2FC12FC1: ("zfs", False)
}


class Fs(Plugin):
    """The "fs" plugin provides information on the filesystem that a file is located in.
    """

    speed = Speed.FAST

    attributes = [
        ("type", String, "The name of the type of file system."),
        ("remote", Boolean, "Whether the file system is a remote file system.")
    ]

    def get_fs_type(self, path):
        """Return the name of the filesystem for path and whether it is a remote filesystem or not.
        """
        # Create a buffer with approximately the size of the statfs structure, we only care about
        # the first field anyway.
        statfs_struct = (c_uint * 64)()

        path_buf = create_string_buffer(os.fsencode(path))

        ret = libc.statfs64(path_buf, byref(statfs_struct))
        if ret == -1:
            errno = get_errno()
            raise OSError(errno, os.strerror(errno), path)

        f_type = statfs_struct[0]
        return fs_types.get(f_type, ("unknown", False))

    def can_handle(self, entry):
        return True

    def process(self, entry, cached):
        if entry.is_dir():
            path = os.path.abspath(entry.path)
        else:
            path = os.path.abspath(entry.dir)

        name, remote = self.get_fs_type(path)

        yield "type", name
        yield "remote", remote



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

from libff.plugin import *


class Test(Plugin):
    """The "test" plugin offers some tests.
    """

    speed = Speed.SLOW

    use_cache = True

    attributes = [
        ("foo", String, "Some test data.")
    ]

    def can_handle(self, entry):
        return True

    def process(self, entry, cached):
        # pylint:disable=unspecified-encoding
        with open("/nonexistent"):
            pass



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import queue
import signal
import traceback
import multiprocessing

from . import TIMEOUT, Entries, BaseClass, Directory
from .path import join
from .entry import Entry
from .exceptions import EX_PROCESS


class FilesystemWalker(BaseClass):
    """Walk through the directories of a filesystem using multiple processes, collect information
       about the entries that were found and process them accordingly.
    """

    def __init__(self, context):
        super().__init__(context)

        self.processes = []
        self.queue = multiprocessing.Queue()

    def close(self):
        """Close the FilesystemWalker and clean up.
        """
        timeout = 10
        for process in self.processes:
            process.join(timeout)
            if process.exitcode is None:
                # If join() hits the timeout once we don't wait for the other processes to
                # terminate.
                timeout = 0

        self.queue.close()

    def put(self, chunk):
        """Put a chunk of objects in the queue for processing by the FilesystemWalker. Either these
           objects are Directory objects for searching or lists of arguments from an
           ImmediateExecProcessing processor to run as subprocesses in the context of the
           FilesystemWalker.
        """
        self.queue.put(chunk)

    def start_processes(self):
        """Start a pool of processes that walk through the filesystem.
        """
        # We set the processes daemon=True because this seems to make them more reactive on
        # KeyboardInterrupt.
        for index in range(self.args.jobs):
            process = multiprocessing.Process(target=self.loop, args=(index,))
            process.daemon = True
            process.start()
            self.processes.append(process)

    def loop(self, index):
        """Get arguments from the queue and process them until searching has finished.
        """
        # pylint:disable=too-many-branches,attribute-defined-outside-init,broad-except
        if __debug__:
            self.index = index
            self.logger.debug_proc(self.index, "started")

        if not (__debug__ and self.args.profile):
            # When profiling, loop() runs in the main thread, so we have to allow
            # KeyboardInterrupt.
            signal.signal(signal.SIGINT, signal.SIG_IGN)

        try:
            while not self.context.is_stopping():
                try:
                    objs = self.queue.get(timeout=TIMEOUT)
                except queue.Empty:
                    if self.context.barrier_wait():
                        break
                    else:
                        continue

                if isinstance(objs, Entries):
                    if __debug__:
                        self.logger.debug_proc(self.index,
                                f"got Entries object with {len(objs.entries)} entries "\
                                f"from {objs.parent.start.root}")

                    self.process_entries(*objs)

                elif isinstance(objs[0], Directory):
                    if __debug__:
                        self.logger.debug_proc(self.index, f"got {len(objs)} Directory entries")

                    for obj in objs:
                        self.process_directory(obj)

                else:
                    if __debug__:
                        self.logger.debug_proc(self.index, f"got {len(objs)} process arguments")

                    for obj in objs:
                        self.process_arguments(obj)

            self.cache.close()

        except Exception:
            # Terminate all processes if an unexpected error occurs.
            self.context.set_exitcode(EX_PROCESS)
            self.context.stop()

            # It proves to be safer to print the traceback ourselves instead of relying on
            # multiprocessing to do it for us.
            traceback.print_exc()

        if __debug__:
            self.logger.debug_proc(self.index, "stopped")

    def scan_directory(self, parent):
        """Scan the directory `parent` and produce a list of entries and the names of .*ignore
           files that were found.
        """
        entries = []
        ignore_paths = parent.ignore_paths.copy()

        try:
            with os.scandir(join(parent.start.root, parent.relpath)) as direntries:
                for direntry in direntries:
                    try:
                        status = direntry.stat(follow_symlinks=self.args.follow_symlinks)
                        entry = Entry(parent.start, join(parent.relpath, direntry.name),
                                status, ignore_paths)

                    except FileNotFoundError:
                        # Do not warn about files that have vanished.
                        continue
                    except PermissionError:
                        # Do not warn about files that we are not allowed to access.
                        continue
                    except OSError as exc:
                        self.logger.warning(exc)
                        continue

                    entries.append(entry)

                    if entry.name in self.args.ignore_files:
                        if __debug__:
                            self.logger.debug("walk", f"Found ignore file {entry.name!r} "\
                                    f"in {entry.dirname!r}")
                        # Please note that we take advantage of the side-effect, that we can still
                        # update the same ignore_paths list we already passed to a number of Entry
                        # objects earlier.
                        ignore_paths.append(entry.abspath)

        except FileNotFoundError:
            # Do not warn about directories that have vanished.
            pass
        except PermissionError:
            # Do not warn about directories that we are not allowed to enter.
            pass
        except OSError as exc:
            self.logger.warning(exc)

        return entries

    def process_directory(self, parent):
        """Scan the directory `parent` for entries, collect .*ignore files, and process the entries
           or distribute them to other FilesystemWalkers for processing.
        """
        entries = self.scan_directory(parent)
        if entries:
            self.process_entries(parent, entries)

    def process_entries(self, parent, entries):
        """Go through the list of entries and see which ones we have to ignore, and exclude, which
           ones match and will be taken to further processing and which directory entries to
           descend into.
        """
        search = []
        process = []

        try:
            while entries:
                entry = entries.pop(0)
                is_dir = entry.is_dir()

                if self.excluder.test(entry):
                    continue

                elif self.matcher.test(entry):
                    process.append(entry)

                if is_dir:
                    search.append(Directory(parent.start, entry.relpath, entry.ignore_paths))

                # If there are plugins active that are marked as slow, this means that processing
                # entries may take more time than usual Here we share the load of this current
                # process with other processes that are idle at the moment.
                if self.registry.optimize_for_slow_plugins and \
                        self.context.idle_processes() / self.args.jobs > 0.25 and \
                        len(entries) > 10:
                    split = len(entries) // 2 + 1
                    self.queue.put(Entries(parent, entries[split:]))
                    entries = entries[:split]

        except OSError as exc:
            self.logger.warning(exc)

        if search:
            # Calculate a reasonable size for a chunk of Entry objects based on the number of
            # parallel jobs, use at least 10 and at most 100 entries to avoid unnecessary overhead.
            chunk_size = min(100, max(10, int(len(search) / self.args.jobs) + 1))

            for i in range(0, len(search), chunk_size):
                self.queue.put(search[i:i + chunk_size])

        while process:
            # Reduce number of Entry object that are sent in one go. This resolves issues with
            # directories that contain a large number of entries.
            self.processing.process(process[:100])
            process = process[100:]

    def process_arguments(self, arguments):
        """Call a subprocess with a list of arguments and wait for its completion.
        """
        self.context.run_exec_process(arguments)



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import re
import stat
from enum import Enum

from .convert import parse_size, parse_time, format_size, format_time, \
    parse_duration, format_duration


class Count(Enum):
    """Specify if a value shall be summed up (TOTAL), counted individually (COUNT) or raise an
       error (UNCOUNTABLE).
    """
    COUNT = 0
    TOTAL = 1
    UNCOUNTABLE = 2


class Type:
    """Type class that implements how an attribute interfaces with the outside world.
    """

    # A set of operators that this Type supports. There is a fixed number of available operators.
    # The first group is more suited for strings the second group more for numbers.
    #
    #    =          equals
    #    :          contains
    #    ~          regex match
    #    %          glob match
    #
    #    +=         greater equal
    #    +          greater
    #    -=         less equal
    #    -          less
    #
    operators = ()

    # This value is used as a fallback value during sorting, when a Plugin does not provide a
    # particular attribute. It must be suitable for using it in a comparison to another value of
    # this Type, e.g. the empty string for strings or 0 for integers.
    sort_none = None

    # Return how this attribute and value shall be included in the statistics. If the value itself
    # shall be summed up to a total (TOTAL) or whether each individual value increases a count by 1
    # (COUNT). TOTAL is most suited for things like size or duration whose values vary widely but
    # are interesting in total, COUNT is best for string values like e.g. extension, type etc. to
    # count how many files have a certain extension or a certain type. If a type is uncountable it
    # must use UNCOUNTABLE.
    count = Count.COUNT

    # Indicate whether this Type is a string type. This affects case conversion, see test().
    string_type = False

    # Limit test values/patterns to certain choices. Should be a set of values that are allowed.
    choices = None

    def __init_subclass__(cls):
        super().__init_subclass__()

        cls.name = cls.__name__.lower()

    @classmethod
    def input(cls, value):
        """Convert a value from a string from the command-line to the required type.
        """
        return value

    @classmethod
    def test(cls, entry, test, value):
        """Return True or False depending on whether the `test` matches `value`. `test` is a
           namedtuple with the following attributes:

                attribute           An Attribute(plugin, name) namedtuple.
                operator            The operator: '=', '~', etc.
                type                The Type class of the value.
                value               The value/pattern to test `value` against.
                ignore_case         Whether or not to consider case for string comparison,
                                    None if the Type is not a string type.

            If both string_type and `test.ignore_case` are True test.value will be in lowercase. It
            is up to you to convert `value` to lowercase as well.
        """
        raise NotImplementedError

    @classmethod
    def output(cls, args, modifier, value):
        # pylint:disable=unused-argument
        """Create a representation of the value for output on the screen. `args` is the global
           ArgumentParser object. At the moment it is only used to detect whether --si was
           specified or not. Modifier is either "h" for human readable, "x" for hexadecimal, "o"
           for octal or None.
        """
        return str(value)

    @classmethod
    def check_type(cls, value):
        """Check if a value conforms to this Type class. This is used to check whether values
           returned by plugins have the correct type and is only used in __debug__ mode. The
           default implementation uses the type of the .sort_none attribute, but this may be
           insufficient, e.g. for compound types like lists.
        """
        return isinstance(value, type(cls.sort_none))

    @classmethod
    def sort_key(cls, value):
        """Extract a comparison key from `value` that will be used for sorting.
        """
        return value


class String(Type):
    """Simple string. Examples: 'file.ext=txt' 'mime.mime~py'
    """

    operators = ("=", ":", "~", "%")
    sort_none = ""
    count = Count.COUNT
    string_type = True

    @classmethod
    def test(cls, entry, test, value):
        # If the test is case insensitive, the test.value argument has already been converted to
        # lower case.
        if test.ignore_case:
            value = value.lower()

        if test.operator == "%":
            if test.value.match(value, value, False):
                return test.value.include
            else:
                return not test.value.include

        return test.operator == "=" and value == test.value or \
                test.operator == ":" and test.value in value or \
                test.operator == "~" and test.value.search(value) is not None

    @classmethod
    def sort_key(cls, value):
        return value.lower()


class Path(String):
    """String containing a filesystem path or a portion thereof. Examples: 'file.name%*.txt'
       'file.dir=foo/bar'
    """

    @classmethod
    def test(cls, entry, test, value):
        if test.ignore_case:
            value = value.lower()

        if test.operator == "%":
            basename = entry.name.lower() if test.ignore_case else entry.name
            is_dir = entry.is_dir()

            if test.value.match(value, basename, is_dir):
                return test.value.include
            else:
                return not test.value.include

        return super().test(entry, test, value)

    @classmethod
    def sort_key(cls, value):
        value = value.lower()
        return os.sep.join(v.lstrip(".") for v in value.split(os.sep))


class FileType(Type):
    """String representing the type of a file. The value is one of d, directory, f, file, l,
       symlink, s, socket, p, pipe, fifo, char, block, door, port, whiteout or other. Example:
       'type=f'
    """

    operators = ("=",)
    sort_none = ""
    count = Count.COUNT
    choices = set(["d", "directory", "f", "file", "l", "symlink", "s", "socket", "p", "pipe",
        "fifo", "char", "block", "door", "port", "whiteout", "other"])

    @classmethod
    def test(cls, entry, test, value):
        test_value = test.value.lower()
        if test_value == "d":
            test_value = "directory"
        elif test_value == "f":
            test_value = "file"
        elif test_value == "l":
            test_value = "symlink"
        elif test_value == "s":
            test_value = "socket"
        elif test_value in ("p", "pipe"): # compatible to fd.
            test_value = "fifo"
        return test.operator == "=" and value == test_value


class ListOfStrings(Type):
    """List of strings, e.g. the lines of a file or filenames inside a tar archive. A given test
       will be applied to each string in the list. Example: 'grep.lines~foobar',
       'tar.members=./foobar.txt'
    """

    operators = String.operators
    sort_none = []
    count = Count.UNCOUNTABLE
    string_type = True
    name = String.name + "[]"

    @classmethod
    def test(cls, entry, test, value):
        return any(String.test(entry, test, v) for v in value)

    @classmethod
    def output(cls, args, modifier, value):
        return ",".join(value)

    @classmethod
    def check_type(cls, value):
        if not isinstance(value, (list, set)):
            return False
        return all(isinstance(v, type(String.sort_none)) for v in value)


class Number(Type):
    """Simple number. Examples: 'file.uid=1000' 'file.links+=2'
    """

    operators = ("+=", "-=", "+", "-", "=")
    sort_none = 0

    @classmethod
    def input(cls, value):
        try:
            value = int(value)
        except ValueError as exc:
            raise ValueError(f"unable to parse {value!r}") from exc
        if value < 0:
            raise ValueError("value must not be less than 0")
        return value

    @classmethod
    def output(cls, args, modifier, value):
        if modifier == "x":
            return f"{value:x}"
        elif modifier == "o":
            return f"{value:o}"
        else:
            return super().output(args, modifier, value)

    @classmethod
    def test(cls, entry, test, value):
        return test.operator == "=" and value == test.value or \
                test.operator == "+=" and value >= test.value or \
                test.operator == "-=" and value <= test.value or \
                test.operator == "+" and value > test.value or \
                test.operator == "-" and value < test.value


class Mode(Type):
    """File permissions. Either octal numbers or symbolic notation is supported. Examples:
       'file.perm=a+rwx' 'file.perm-u+x' 'file.perm+700'
    """

    operators = ("=", ":", "~")
    sort_none = 0
    count = Count.COUNT

    regex = re.compile(r"([ugoa]*)([-+=])([rwxXst]+)|([-+=])?([0-7]+)$")

    MODE_ALL = stat.S_ISUID | stat.S_ISGID | stat.S_ISVTX | \
               stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO

    @classmethod
    def parse_mode(cls, value):
        """Parse an octal or symbolic file mode.
        """
        # pylint:disable=too-many-branches
        affected = 0
        result = 0
        for part in value.split(","):
            val = 0

            match = cls.regex.match(part)
            if match is None:
                raise ValueError(f"unable to parse {value!r}")

            if match.group(1) is not None:
                affected_str, operator, value_str = match.groups()[:3]

                if not affected_str:
                    affected_str = "a"

                for char in affected_str:
                    if char == "u":
                        affected |= stat.S_ISUID | stat.S_IRWXU
                    elif char == "g":
                        affected |= stat.S_ISGID | stat.S_IRWXG
                    elif char == "o":
                        affected |= stat.S_ISVTX | stat.S_IRWXO
                    elif char == "a":
                        affected |= cls.MODE_ALL

                if value_str == "u":
                    val |= stat.S_IRWXU
                elif value_str == "g":
                    val |= stat.S_IRWXG
                elif value_str == "o":
                    val |= stat.S_IRWXO
                else:
                    for char in value_str:
                        if char == "r":
                            val |= stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH
                        elif char == "w":
                            val |= stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH
                        elif char == "x":
                            val |= stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
                        elif char == "s":
                            val |= stat.S_ISUID | stat.S_ISGID
                        elif char == "t":
                            val |= stat.S_ISVTX

            else:
                operator, val = match.groups()[3:]
                if not operator:
                    operator = "="
                affected = cls.MODE_ALL
                val = int(val, 8)

            if operator in "+=":
                result |= val
            elif operator == "-":
                result &= ~val

        return affected & result

    @classmethod
    def input(cls, value):
        return cls.parse_mode(value)

    @classmethod
    def output(cls, args, modifier, value):
        if modifier == "h":
            return stat.filemode(value)
        elif modifier == "x":
            return f"{value:x}"
        elif modifier == "o":
            return f"{value:o}"
        else:
            return super().output(args, modifier, value)

    @classmethod
    def test(cls, entry, test, value):
        test_value = test.value & cls.MODE_ALL
        value = value & cls.MODE_ALL
        return test.operator == "=" and value == test_value or \
               test.operator == ":" and value & test_value == test_value or \
               test.operator == "~" and value & test_value != 0


class Size(Number):
    """File size. Example: 'file.size+=100M'
    """

    count = Count.TOTAL

    @classmethod
    def input(cls, value):
        return parse_size(value)

    @classmethod
    def output(cls, args, modifier, value):
        if modifier == "h":
            return format_size(value, base=1000 if args.si else 1024)
        else:
            return super().output(args, modifier, value)


class Time(Number):
    """Time in seconds since 1970-01-01 00:00:00. Examples: 'file.time=1589234400'
       'file.time+=2020-05-12' 'file.time+=1d12h'
    """

    @classmethod
    def input(cls, value):
        return parse_time(value)

    @classmethod
    def output(cls, args, modifier, value):
        if modifier == "h":
            return format_time(value)
        else:
            return super().output(args, modifier, value)


class Duration(Number):
    """Duration in seconds. Example: 'medium.duration+=1h30m'
    """

    count = Count.TOTAL

    @classmethod
    def input(cls, value):
        return parse_duration(value)

    @classmethod
    def output(cls, args, modifier, value):
        if modifier == "h":
            return format_duration(value)
        else:
            return super().output(args, modifier, value)


class Boolean(Type):
    """Boolean value. May be one of (true, t, 1, yes, y, on) or (false, f, 0, no, n, off). The case
       is ignored. Example: 'empty=yes'
    """

    operators = ("=",)
    sort_none = False
    count = Count.COUNT

    true = ("true", "t", "1", "yes", "y", "on")
    false = ("false", "f", "0", "no", "n", "off")

    @classmethod
    def input(cls, value):
        assert value.lower() in cls.true + cls.false
        return value.lower() in cls.true

    @classmethod
    def output(cls, args, modifier, value):
        return str(value).lower()

    @classmethod
    def test(cls, entry, test, value):
        return test.operator == "=" and value is test.value



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import subprocess

from libff.plugin import *


class Pacman(Plugin):
    """The "pacman" plugin provides information about which files were installed using the package
       manager pacman on an Arch Linux. It requires the 'pacman' executable.
    """

    attributes = [
        ("installed", Boolean, "Whether the file belongs to an Arch Linux system package "\
                               "installed by the package manager pacman."),
        ("pkgname", String, "Name of package the file belongs to.")
    ]

    def setup(self):
        self.pkglist = {}
        self.filelist = set()
        with subprocess.Popen(["pacman", "-Ql"], stdout=subprocess.PIPE, text=True) as proc:
            for line in proc.stdout:
                line = line.strip()
                pkgname, path = line.split(None, 1)
                path = path.rstrip(os.sep)
                self.filelist.add(path)
                self.pkglist.setdefault(pkgname, set()).add(path)

    def can_handle(self, entry):
        return True

    def process(self, entry, cached):
        if entry.path in self.filelist:
            yield "installed", True
            for pkgname, filelist in self.pkglist.items():
                if entry.path in filelist:
                    yield "pkgname", pkgname
        else:
            yield "installed", False



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

from libff.path import split
from libff.ignore import GitIgnore
from libff.plugin import *


class Ignore(Plugin):
    """The "ignore" plugin shows whether files match patterns from .{,git,fd,ff}ignore files. It is
       useful e.g. inside working copies of git repositories.
    """

    speed = Speed.SLOW

    attributes = [
        ("ignored", Boolean, "Whether the file matches patterns in a .{,git,fd,ff}ignore file."),
        ("path", String, "The path to the ignore file that contained the matching pattern."),
    ]

    def __init__(self):
        super().__init__()

        self.ignore_cache = {}
        self.ignores_cache = {}

    def can_handle(self, entry):
        return True

    def get_ignores(self, paths):
        """Return a list of GitIgnore objects that apply to this particular directory.
        """
        paths = tuple(paths)

        # Construct a chain of GitIgnore objects one for each ignore file found from the parent
        # directories on down to the current directory.
        if paths not in self.ignores_cache:
            ignores = []
            for path in paths:
                # Reuse GitIgnore objects that have already been instantiated.
                try:
                    ignore = self.ignore_cache[path]
                except KeyError:
                    try:
                        ignore = GitIgnore(*split(path))
                    except OSError as exc:
                        self.logger.warning(exc, tag=f"ignore-{path}")
                        continue

                    self.ignore_cache[path] = ignore

                ignores.append(ignore)

            self.ignores_cache[paths] = ignores

        return self.ignores_cache[paths]

    def is_ignored(self, entry):
        """Return True if the Entry object matches a pattern in one of the ignore files that apply.
        """
        if entry.ignore_paths:
            ignores = self.get_ignores(entry.ignore_paths)
            return GitIgnore.match(ignores, entry.abspath, entry.name, entry.is_dir())
        else:
            return False, None

    def process(self, entry, cached):
        ignored, path = self.is_ignored(entry)
        return {"ignored": ignored, "path": path if path is not None else ""}



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import threading
import subprocess
import multiprocessing
import multiprocessing.sharedctypes

from . import TIMEOUT
from .exceptions import EX_SUBPROCESS


class Context:
    """The Context class contains a global state and is globally accessible so that all components
       find all other components.
    """
    # pylint:disable=too-many-instance-attributes

    warnings_emitted = set()

    def __init__(self):
        # These are added later.
        self.args = None
        self.logger = None
        self.cache = None
        self.registry = None
        self.matcher = None
        self.excluder = None
        self.walker = None
        self.processing = None

        self.global_lock = None
        self.stop_queue = None
        self.barrier = None
        self.exitcode_object = None
        self.process_succeeded = None
        self.process_failed = None

        self.cache_hits = None
        self.cache_misses = None

    def setup(self):
        """Set up multiprocessing and inter-process communication.
        """
        self.global_lock = multiprocessing.Lock()

        # We use a Queue object to emulate a stop event, because multiprocessing.Event did not work
        # the way I expected it.
        self.stop_queue = multiprocessing.Queue(self.args.jobs)

        # Create a Barrier with the number of processes plus one for the main process.
        if __debug__ and self.args.profile:
            num_jobs = 1
        else:
            num_jobs = self.args.jobs + 1
        self.barrier = multiprocessing.Barrier(num_jobs)

        # The exitcode object provides a way for --exec and --exec-batch processes to feed back
        # errors to the main process.
        self.exitcode_object = multiprocessing.sharedctypes.RawValue("i", 0)
        self.process_succeeded = multiprocessing.Value("L")
        self.process_failed = multiprocessing.Value("L")

        # Provide a shared counter for cache hits and misses.
        self.cache_hits = multiprocessing.Value("L")
        self.cache_misses = multiprocessing.Value("L")

    def close(self):
        """Close the Context and clean up.
        """
        self.stop_queue.close()

    def set_exitcode(self, exitcode):
        """Set the shared resource 'exitcode' to the value that will be used as the main exit code.
        """
        self.exitcode_object.value = exitcode

    @property
    def exitcode(self):
        """The exitcode to use as the main exit code.
        """
        return self.exitcode_object.value

    def barrier_wait(self, timeout=TIMEOUT):
        """Wait for all processes to reach the Barrier. If all processes do that means that there
           is no more input and processing is finished.
        """
        try:
            self.barrier.wait(timeout)
        except threading.BrokenBarrierError:
            self.barrier.reset()
            return self.is_stopping()
        else:
            return True

    def idle_processes(self):
        """Return the number of idle processes, i.e. processes that are waiting for the barrier
           because there is nothing in the queue.
        """
        return self.barrier.n_waiting

    def is_stopping(self):
        """Return True all Processes are supposed to stop.
        """
        return self.stop_queue.qsize() > 0

    def stop(self):
        """Stop all processes immediately.
        """
        if not self.is_stopping():
            self.stop_queue.put(None)

    def run_exec_process(self, args):
        """Run a subprocess with the arguments from --exec. If the process fails, terminate the
           main loop and all other running subprocesses depending on the value of the --halt
           option.
        """
        exitcode = 1

        try:
            # pylint:disable=consider-using-with
            process = subprocess.Popen(args)

            if self.args.halt in ("never", "soon"):
                # Let this process finish regardless of the state of the main loop.
                exitcode = process.wait()

            elif self.args.halt == "now":
                # If the main loop is stopped, stop this process as well.
                while not self.is_stopping():
                    try:
                        exitcode = process.wait(0.1)
                        break
                    except subprocess.TimeoutExpired:
                        continue

                else:
                    process.terminate()

        except OSError as exc:
            self.logger.error(exc, None)

        if exitcode != 0:
            self.set_exitcode(EX_SUBPROCESS)
            with self.process_failed.get_lock():
                self.process_failed.value += 1

            # Stop the main loop (and all running subprocesses).
            if self.args.halt in ("soon", "now"):
                self.stop()

        else:
            with self.process_succeeded.get_lock():
                self.process_succeeded.value += 1



#!/usr/bin/python3
# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import sys
import json
import time
import queue
import threading
import subprocess
import collections
import multiprocessing

from . import TIMEOUT, BaseClass
from .type import Count
from .console import Console, JsonConsole, NullConsole, ColorConsole, \
    JsonlConsole
from .exceptions import EX_PROCESS, EX_SUBPROCESS, SubprocessError


class Parallel(BaseClass):
    """Implement a pool of threads each running a subprocess. Normally -x/--exec commands are
       issued directly in FilesystemWalker, but when entries are collated first (e.g. by --sort) we
       use this pool afterwards to call -x/--exec commands in parallel.
    """

    def __init__(self, context):
        super().__init__(context)

        self.queue = queue.Queue()

        self.threads = []
        for _ in range(self.args.jobs):
            thread = threading.Thread(target=self.thread)
            thread.daemon = True
            thread.start()
            self.threads.append(thread)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        for thread in self.threads:
            self.queue.put(None)
        for thread in self.threads:
            thread.join()

    def thread(self):
        """Loop over a queue and execute subprocesses.
        """
        while not self.context.is_stopping():
            try:
                args = self.queue.get(timeout=TIMEOUT)
            except queue.Empty:
                continue

            if args is None:
                break

            self.context.run_exec_process(args)

    def add_job(self, entry):
        """Add a subprocess to the queue for an Entry object.
        """
        try:
            self.queue.put(self.args.exec.render(entry))
        except KeyError:
            pass


class BaseProcessing(BaseClass):
    """Base class for processing Entry objects.
    """

    def __init__(self, context):
        super().__init__(context)

        self.next_processes_check = time.time() + 10

    def check_for_failed_processes(self):
        """Check regularly if all processes are still active and allow an ordered shutdown even if
           processes terminated abnormally. In this case, abnormal shutdown means that a process
           crashed in a way that it could not do the usual cleanup procedure (print a traceback,
           send stop signal to all other processes, set the exitcode). This may happen if one the
           process is terminated by a signal e.g. a segmentation fault.
        """

        if time.time() >= self.next_processes_check:
            for index, process in enumerate(self.walker.processes):
                if not process.is_alive():
                    self.logger.warning(f"process #{index} terminated abnormally")
                    self.context.set_exitcode(EX_PROCESS)
                    self.context.stop()
                    return True

            self.next_processes_check = time.time() + 10

        return False

    def processing_done(self, timeout=TIMEOUT):
        """Check if processing is done (because all processes wait on the barrier) or one or more
           processes failed unexpectedly.
        """
        return self.context.barrier_wait(timeout) or self.check_for_failed_processes()

    def process(self, entries):
        """Process a chunk of Entry objects.
        """
        raise NotImplementedError

    def loop(self):
        """Wait for the Walker processes to finish (while possibly collecting entries).
        """

    def finalize(self):
        """Process the collected list of Entry objects (in case this is a Processing class that
           collects entries first).
        """

    def close(self):
        """Allow doing some clean up.
        """


class ImmediateBaseProcessing(BaseProcessing):
    """Base class for Processing classes that will process entries without collecting them first.
    """
    # pylint:disable=abstract-method

    def loop(self):
        while not self.processing_done(1):
            pass


class BaseConsoleProcessing(ImmediateBaseProcessing):
    """Processing base class for console output.
    """

    def __init__(self, context):
        super().__init__(context)

        if self.args.count is not None:
            console_cls = NullConsole

        elif self.args.json == "json":
            console_cls = JsonConsole

        elif self.args.json == "jsonl":
            console_cls = JsonlConsole

        elif sys.stdout.isatty() if self.args.color == "auto" else self.args.color == "always":
            console_cls = ColorConsole

        else:
            console_cls = Console

        self.console = console_cls(self.context)

    def process(self, entries):
        for entry in entries:
            self.console.process(entry)

    def close(self):
        super().close()

        self.console.close()


class ImmediateExecProcessing(ImmediateBaseProcessing):
    """Processing class for direct execution of --exec jobs.

       Ensure an even distribution of jobs to processes, therefore put only one Call in the queue
       at a time.
    """

    def process(self, entries):
        for entry in entries:
            try:
                self.walker.put([self.args.exec.render(entry)])
            except KeyError:
                pass


class ImmediateConsoleProcessing(BaseConsoleProcessing):
    """Processing class for direct output to the console.
    """


class CollectiveMixin(BaseProcessing):
    """Processing mixin class that collects all entries first and then does the processing
       afterwards.
    """

    def __init__(self, context):
        super().__init__(context)

        self.out_queue = multiprocessing.Queue()
        self.entries = []
        self.count = collections.Counter()

    def process(self, entries):
        self.out_queue.put(entries)

    def entries_range(self):
        """Limit the list of entries according to the -l/--limit option.
        """
        if self.args.limit is None:
            return self.entries
        else:
            start, stop = self.args.limit
            return self.entries[start:stop]

    def loop(self):
        while True:
            try:
                entries = self.out_queue.get(timeout=TIMEOUT)
            except queue.Empty:
                if self.processing_done():
                    break
                else:
                    continue

            if self.args.count is not None:
                self.collect_count(entries)

            self.entries += entries

    def finalize(self):
        if self.args.sort is not None:
            self.entries.sort(key=self.args.sort.render, reverse=self.args.reverse)

    def close(self):
        super().close()

        self.out_queue.close()

        if self.args.count is not None:
            self.print_count()

    def collect_count(self, entries):
        """Collect value count.
        """
        for entry in entries:
            # Count the total number of entries found.
            self.count["_total"] += 1

            for field in self.args.count:
                try:
                    value = self.registry.get_attribute(entry, field.attribute)
                except KeyError:
                    continue

                if field.type.count is Count.TOTAL:
                    # Create a total of all the values.
                    self.count[field] += value

                elif field.type.count is Count.COUNT:
                    # Count each value individually.
                    self.count.setdefault(field, collections.Counter())[value] += 1

    def prepare_count(self):
        """Prepare the collected value count for printing.
        """
        # First prepare a plain dictionary structure suited for JSON output.
        count = {}
        for field, value in self.count.items():
            if isinstance(field, str):
                # E.g. "_total".
                count[field] = value

            else:
                name = str(field.attribute)

                if isinstance(value, dict):
                    # This is a dictionary with a by-value count.
                    count[name] = {}
                    for key, val in value.items():
                        # Format the key which is the counted value in this case, JSON objects
                        # allow only string keys anyway.
                        key = field.type.output(self.args, field.modifier, key)
                        count[name][key] = val

                else:
                    if field.modifier is not None:
                        # Format the value only in case there is a modifier, so that numbers won't
                        # turn out as strings.
                        value = field.type.output(self.args, field.modifier, value)
                    count[name] = value

        return count

    def print_count(self):
        """Print collected value count.
        """
        count = self.prepare_count()

        # Print the count to stdout.
        if self.args.json is not None:
            json.dump(count, sys.stdout, sort_keys=True)
            print()

        else:
            for key, value in sorted(count.items()):
                if isinstance(value, dict):
                    for k, v in sorted(value.items(),
                            key=lambda t: int(t[0]) if t[0].isdigit() else t[0]):
                        print(f"{key}[{k}]={v}")
                else:
                    print(f"{key}={value}")


class CollectiveConsoleProcessing(CollectiveMixin, BaseConsoleProcessing):
    """Processing class for collection and ordering of entries with output to the console
       afterwards.
    """

    def finalize(self):
        super().finalize()

        for entry in self.entries_range():
            self.console.process(entry)


class CollectiveExecProcessing(CollectiveMixin, BaseProcessing):
    """Processing class for collective execution of --exec and --exec-batch jobs.
    """

    def finalize(self):
        super().finalize()

        entries = self.entries_range()

        if not entries:
            return

        if self.args.exec is not None:
            with Parallel(self.context) as parallel:
                for entry in entries:
                    parallel.add_job(entry)

        elif self.args.exec_batch is not None:
            args = self.args.exec_batch.render(entries)

            try:
                process = subprocess.run(args, check=False)
                if process.returncode != 0:
                    self.context.set_exitcode(EX_SUBPROCESS)
            except OSError as exc:
                raise SubprocessError(str(exc)) from exc

        else:
            raise AssertionError("wrong usage of CollectiveExecProcessing class")


class ImmediateGenerator(ImmediateBaseProcessing):
    """Processing class that allows immediate iteration over the results.
    """

    def __init__(self, context):
        super().__init__(context)

        self.out_queue = multiprocessing.Queue()

    def loop(self):
        pass

    def finalize(self):
        pass

    def process(self, entries):
        self.out_queue.put(entries)

    def __iter__(self):
        while True:
            try:
                for entry in self.out_queue.get(timeout=TIMEOUT):
                    yield self.args.output.to_dict(entry)
            except queue.Empty:
                if self.processing_done():
                    break
                else:
                    continue


class CollectiveGenerator(ImmediateGenerator):
    """Processing class that allows iteration over the results after they have been postprocessed.
    """

    def __init__(self, context):
        super().__init__(context)

        self.entries = []

    def loop(self):
        while True:
            try:
                entries = self.out_queue.get(timeout=TIMEOUT)
            except queue.Empty:
                if self.processing_done():
                    break
                else:
                    continue

            self.entries += entries

    def finalize(self):
        super().finalize()

        self.entries.sort(key=self.args.sort.render, reverse=self.args.reverse)

    def __iter__(self):
        yield from (self.args.output.to_dict(entry) for entry in self.entries)



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import re
import itertools
import collections

from .type import Count
from .exceptions import UsageError

Field = collections.namedtuple("Field", "attribute type width modifier")


class AttributeDict(dict):
    """Allow attribute access to a dictionary. Dot field separators are replaced with underscores.
    """

    def __init__(self, fields):
        super().__init__()
        self._fields = {field.replace(".", "_"): field for field in fields}

    def __getattr__(self, name):
        if name not in self._fields:
            raise KeyError(name)
        else:
            return self[self._fields[name]]


class Fields(list):
    """A class that stores a list of attributes as fields.
    """

    regex = re.compile(r"^((?:[a-zA-Z][a-zA-Z0-9_]+\.)?[a-zA-Z][a-zA-Z0-9_]+?)"\
            r"(?::(-?\d+)?(h|o|x|n|v)?)?$")

    def __init__(self, context, argument):
        super().__init__()

        self.context = context
        self.args = self.context.args
        self.registry = self.context.registry

        self.store(argument)

    def make_field(self, string):
        """Create a Field from an attribute string.
        """
        match = self.regex.match(string.strip())
        if match is None:
            raise UsageError(f"Invalid attribute {string!r}")
        name, width, modifier = match.groups()
        attribute = self.registry.setup_attribute(name)
        type_cls = self.registry.get_attribute_type(attribute)
        return Field(attribute, type_cls, width, modifier)

    def store(self, argument):
        """Convert argument to a list of Field objects and store them.
        """
        raise NotImplementedError


class OutputFields(Fields):
    """Store a list of fields that help with formatting output.
    """

    def __init__(self, context, argument):
        super().__init__(context, argument)

        self.fields = [str(field.attribute) for field in self]

    def store(self, argument):
        for string in argument:
            self.append(self.make_field(string))

    def to_dict(self, entry):
        """Create a dictionary from the list of fields.
        """
        record = AttributeDict(self.fields)
        for field in self:
            try:
                value, type_cls = \
                        self.context.registry.get_attribute_and_type(entry, field.attribute)
            except KeyError:
                value = None
            else:
                if field.modifier == "h":
                    value = type_cls.output(self.args, field.modifier, value)

            record[str(field.attribute)] = value

        return record


class CountFields(OutputFields):
    """Store a list of fields that help with formatting statistics output.
    """

    def store(self, argument):
        super().store(argument)

        for field in self:
            if field.type.count is Count.UNCOUNTABLE:
                raise UsageError(f"Attribute {field.attribute} is not suited for --count!")


class SortFields(OutputFields):
    """Store a list of fields that help with sorting Entry objects.
    """

    regex_digits = re.compile(r"(\d+)")

    def prepare_component(self, component):
        """Prepare a path component to act as a part of a sort key.
        """
        # Strip away all file extensions that don't contain digits.
        extensions = []
        while component:
            comp, ext = os.path.splitext(component)
            ext = ext.lstrip(os.extsep)
            if not ext or ext[0].isdigit():
                break
            component = comp
            extensions.insert(0, ext)

        # Split the component into string and number parts. Make sure the list starts with a
        # string.
        parts = self.regex_digits.split(component)
        if parts[0].isdigit():
            parts.insert(0, "")

        # Yield a list of tokens with all numbers converted to ints.
        yield [int(p) if p.isdigit() else p for p in parts]

        # Yield extensions separately.
        if extensions:
            yield extensions

    def prepare_sort_key(self, value):
        """Prepare a sort key from a path for natural/version sorting.
        """
        # Split a full path in its components and prepare each component separately. This should
        # produce output similar to natsort.os_sort_key().
        cmp_key = []
        for component in os.path.normpath(value).split(os.sep):
            cmp_key += list(self.prepare_component(component))
        return cmp_key

    def render(self, entry):
        """Create a sort key from an Entry object.
        """
        output = []
        for field in self:
            try:
                value = self.registry.get_attribute(entry, field.attribute)
            except KeyError:
                output.append(field.type.sort_none)
            else:
                value = field.type.sort_key(value)
                if field.modifier == "v":
                    value = self.prepare_sort_key(value)
                output.append(value)
        return tuple(output)


class ExecFields(Fields):
    """Store a list of fields that help with calling subprocesses.
    """

    regex_placeholder = re.compile(r"({{|}}|{[^}]*})")

    def replace(self, part):
        """Replace a placeholder with a Field.
        """
        if part == "{{":
            field = "{"
        elif part == "}}":
            field = "}"
        elif part == "{}":
            field = self.make_field("file.path")
        elif part == "{/}":
            field = self.make_field("file.name")
        elif part == "{//}":
            field = self.make_field("file.dir")
        elif part == "{.}":
            field = self.make_field("file.pathx")
        elif part == "{/.}":
            field = self.make_field("file.namex")
        elif part == "{..}":
            field = self.make_field("file.ext")
        elif part.startswith("{") and part.endswith("}"):
            field = self.make_field(part[1:-1])
        else:
            field = part
        return field

    def store(self, argument):
        # This list of arguments:
        #
        #    ["convert", "{}", "new-{.}.jpg"]
        #
        # will be translated to:
        #
        #    [["convert"], [Field("file.path")], ["new-", Field("file.pathx"), ".jpg"]]
        #
        for string in argument:
            self.append([self.replace(part)
                for part in self.regex_placeholder.split(string)
                if part])

        # Append a path field if there was no placeholder found in the argument list.
        if not any(isinstance(field, Field) for field in itertools.chain(*self[1:])):
            self.append([self.make_field("file.path")])

    def render_field(self, entry, field):
        """Return a rendered Field value for the Entry object.
        """
        if isinstance(field, Field):
            try:
                value = self.registry.get_attribute(entry, field.attribute)
            except KeyError:
                if self.args.all or field.modifier == "n":
                    return ""
                else:
                    raise
            else:
                return field.type.output(self.args, field.modifier, value)
        else:
            # Fields that do not contain a placeholder will be left unchanged.
            return field

    def render_fields(self, fields, entries, batch):
        """Return a list of arguments with all placeholders replaced.
        """
        output = []
        for subfields in fields:
            # Each element of the list of 'fields' is a alternating sequence of strings and Field
            # objects that act as placeholders, see the comment in store().
            # If in batch mode each Entry from the 'entries' list will produce a new argument in
            # the output argument list, i.e. ["echo", "{/}"] with ["foo", "bar", "baz"] as entries
            # will produce ["echo", "foo", "bar", "baz"].
            # When not in batch mode the same will produce: ["echo", "foo"], ["echo", "bar"],
            # ["echo", "baz"].
            if any(isinstance(a, Field) for a in subfields):
                for entry in entries:
                    # Build one argument for each Entry at a time.
                    try:
                        argument = []
                        for field in subfields:
                            argument.append(self.render_field(entry, field))
                        output.append("".join(argument))

                    except KeyError:
                        # One of the attributes produced no value. If in batch mode, we ignore this
                        # and simply add no new argument. If not in batch mode we want this problem
                        # to be propagated to the caller, so that there is no --exec call.
                        if not batch:
                            raise

            else:
                output.append("".join(subfields))

        return output

    def render(self, entry):
        """Create a list of arguments from an Entry object for calling a subprocess.
        """
        return self.render_fields(self, [entry], False)


class ExecBatchFields(ExecFields):
    """Store a list of fields that help with calling one subprocess with arguments from multiple
       Entry objects.
    """

    def store(self, argument):
        super().store(argument)

        if any(isinstance(field, Field) for field in self[0]):
            raise UsageError("The first part of the command must not contain placeholders!")

    # pylint:disable=arguments-renamed
    def render(self, entries):
        """Create a list of arguments from multiple Entry objects for calling a subprocess.
        """
        return ["".join(self[0])] + self.render_fields(self[1:], entries, True)
